package org.heatdeath.freewiki;

import java.util.*;

public class CachedWikiManager implements WikiManager {
    private WikiManager baseManager;
    private HashMap cache = new HashMap();

    private HashMap wikiCache(String wikiName) {
	if (cache.get(wikiName.toLowerCase()) == null) {
	    cache.put(wikiName.toLowerCase(), new HashMap());
	}
        return (HashMap) cache.get(wikiName.toLowerCase());
    }

    private WikiNodeData getCachedData(String wikiName, String nodeName)
	    throws WikiStoreException {
	if (wikiCache(wikiName).get(nodeName.toLowerCase()) == null) {
	    return null;
	} else {
	    Object o = (WikiNodeData) wikiCache(wikiName).get(nodeName.toLowerCase());
	    if (o instanceof WikiNodeData) {
		return (WikiNodeData) o;
	    } else {
		throw new WikiStoreException((String) o);
	    }
	}
    }

    private synchronized void updateCache(String wikiName, String nodeName,
					  Object data) {
	if (data instanceof WikiNodeData) {
	    WikiNodeData newData = (WikiNodeData) data;
	    WikiNodeData curData;
	    try {
		curData = getCachedData(wikiName, nodeName);
	    } catch (WikiStoreException ex) {
		curData = null;
	    }
	    if (curData == null ||
	            newData.getRevisionNumber() > curData.getRevisionNumber()) {
		wikiCache(wikiName).put(nodeName.toLowerCase(), newData);
		notifyAll();
	    }
	} else {
	    wikiCache(wikiName).put(nodeName.toLowerCase(), data);
	    notifyAll();
	}
    }

    public CachedWikiManager(WikiManager baseManager) {
	this.baseManager = baseManager;
    }

    public void refreshNode(String wikiName, String nodeName) {
	try {
	    WikiNodeData data = baseManager.getNode(wikiName, nodeName);
	    updateCache(wikiName, nodeName, data);
	} catch (WikiStoreException ex) {
	    updateCache(wikiName, nodeName, ex.getMessage());
	}
    }

    public synchronized WikiNodeData getNode(final String wikiName,
	    final String nodeName) throws WikiStoreException {
	if (getCachedData(wikiName, nodeName) == null) {
	    new Thread(new Runnable() {public void run() {
		refreshNode(wikiName, nodeName);
	    }}).start();
	    do {
		try {
		    wait();
		} catch (InterruptedException ex) {
		    ex.printStackTrace();
		    throw new RuntimeException(ex.toString());
		}
	    } while (getCachedData(wikiName, nodeName) == null);
	}
	return getCachedData(wikiName, nodeName);
    }

    public WikiNodeData editNode(final String wikiName, final String nodeName,
	    final String newContents, final int mostRecentRevision)
            throws CollisionException, WikiStoreException {
	class EditRunnable implements Runnable {
	    private WikiNodeData newData = null;
	    private boolean collision = false;
	    private String storeMsg = null;

	    public synchronized void run() {
		try {
		    newData = baseManager.editNode(wikiName, nodeName,
		            newContents, mostRecentRevision);
		    updateCache(wikiName, nodeName, newData);
		} catch (CollisionException ex) {
		    collision = true;
		} catch (WikiStoreException ex) {
		    storeMsg = ex.getMessage();
		}
		notify();
	    }

	    public synchronized WikiNodeData getNewData() throws CollisionException,
	            WikiStoreException {
		try {
		    wait();
		} catch (InterruptedException ex) {
		    ex.printStackTrace();
		    throw new RuntimeException(ex.toString());
		}
		if (collision) {
		    throw new CollisionException();
		}
		if (storeMsg != null) {
		    throw new WikiStoreException(storeMsg);
		}
		return newData;
	    }
	}
	if (getNode(wikiName, nodeName).getRevisionNumber() > mostRecentRevision) {
	    throw new CollisionException();
	}
	EditRunnable r = new EditRunnable();
	new Thread(r).start();
	return r.getNewData();
    }
}
