package org.heatdeath.freewiki;

import java.util.*;
import java.io.*;
import Freenet.client.*;
import Freenet.client.events.*;
import Freenet.support.*;

public class FreenetWikiManager implements WikiManager {
    private int htl;
    private ClientFactory clientFactory;

    private class ClientEventListenerImpl implements ClientEventListener {
	private boolean failed = false;
	private boolean collision = false;

	public synchronized void receive(ClientEvent event) {
	    System.out.println("Got event: " + event.getDescription());
	    if (event instanceof StateReachedEvent) {
		int state = ((StateReachedEvent) event).getState();
		switch(((StateReachedEvent) event).getState()) {
    		    case Request.DONE:
			failed = false;
			notifyAll();
			break;
		    case Request.FAILED:
			failed = true;
			notifyAll();
			break;
		}
	    } else if (event instanceof CollisionEvent) {
		collision = true;
	    }
	}

	public synchronized void waitUntilFinished() {
	    try {
		wait();
	    } catch (InterruptedException ex) {
		ex.printStackTrace();
		throw new RuntimeException(ex.toString());
	    }
	}

	public synchronized boolean isFailed() {
	    return failed;
	}

	public synchronized boolean isCollision() {
	    return collision;
	}
    }

    private String nodeKey(String wikiName, String nodeName, int version) {
	return "KSK@freewiki/" + wikiName.toLowerCase() + '/' +
	        nodeName.toLowerCase() + "-revision-" + version;
    }

    private String readContent(FileBucket dataBucket) throws IOException {
	BufferedReader reader = null;
	try {
	    reader = new BufferedReader(new InputStreamReader(
	            dataBucket.getInputStream(), "UTF-8"));
	    StringBuffer sb = new StringBuffer();
	    String line = reader.readLine();
	    while (line != null) {
		sb.append(line);
		sb.append("\n");
		line = reader.readLine();
	    }
	    return sb.toString();
	} finally {
	    if (reader != null) {
		reader.close();
	    }
	}
    }

    private Date parseMetaData(FileBucket metaBucket)
	    throws IOException, WikiStoreException  {
	if (metaBucket.size() == 0) {
	    return null;
	}
	BufferedReader reader = null;
	try {
	    Date date = null;
	    reader = new BufferedReader(new InputStreamReader(
		    metaBucket.getInputStream(), "UTF-8"));
	    String line = reader.readLine();
	    while (line != null) {
		if (line.indexOf('=') == -1) {
		    break;
		}
		String n = line.substring(0, line.indexOf('=')).toLowerCase();
		String v = line.substring(line.indexOf('=') + 1);
		if (n.equals("x-freewiki-node-date")) {
		    date = new Date(Long.parseLong(v));
		} else if (n.equals("content-type")) {
		    if (!v.toLowerCase().equals("text/x-freewiki-node-contents")) {
			throw new WikiStoreException("Incorrect content type: " + v);
		    }
		}
		line = reader.readLine();
	    }
	    return date;
	} finally {
	    if (reader != null) {
		reader.close();
	    }
	}
    }

    private boolean getKey(String key, FileBucket metaBucket,
	    FileBucket dataBucket) throws IOException, WikiStoreException {
	System.out.println("GETTING KEY: " + key);
	try {
	    DataRequest request = new DataRequest(htl, key, metaBucket, dataBucket);
	    ClientEventListenerImpl l = new ClientEventListenerImpl();
	    request.addEventListener(l);
	    clientFactory.obtainClient(request).execute();
	    l.waitUntilFinished();
	    return l.isFailed();
	} catch (java.net.MalformedURLException ex) {
	    throw new WikiStoreException("Cannot get key " + key + ": " + ex);
	}
    }

    private WikiNodeData.RevisionData getRevision(String wikiName, String nodeName,
	    int revisionNumber) throws WikiStoreException {
	FileBucket metaBucket = null;
	FileBucket dataBucket = null;
	try {
	    metaBucket = new FileBucket();
	    dataBucket = new FileBucket();
	    boolean failed = getKey(nodeKey(wikiName, nodeName, revisionNumber),
				    metaBucket, dataBucket);
	    if (failed) {
		return null;
	    }
	    return new WikiNodeData.RevisionData(revisionNumber,
	            parseMetaData(metaBucket), readContent(dataBucket));
	} catch (IOException ex) {
	    throw new WikiStoreException("Cannot get revision: " + ex);
	} finally {
	    cleanupBuckets(dataBucket, metaBucket, null);
	}
    }

    private void cleanupBuckets(FileBucket dataBucket, FileBucket metaBucket,
				FileBucket ctBucket) {
	try {
	    if (dataBucket != null) {
		dataBucket.getFile().delete();
	    }
	} finally {
	    try {
		if (metaBucket != null) {
		    metaBucket.getFile().delete();
		}
	    } finally {
		if (ctBucket != null) {
		    ctBucket.getFile().delete();
		}
	    }
	}
    }

    private WikiNodeData getNodeRevision(String wikiName, String nodeName,
					 int maxRevision)
	    throws WikiStoreException {
	ArrayList revisions = new ArrayList();
	int version = 1;
	while (maxRevision == -1 || version <= maxRevision) {
	    WikiNodeData.RevisionData rev = getRevision(wikiName, nodeName, version);
	    if (rev == null) {
		break;
	    }
	    revisions.add(rev);
	    version++;
	}
	return new WikiNodeData(wikiName, nodeName,
				Collections.unmodifiableList(revisions));
    }
    
    private void writeContents(String newContents, FileBucket dataBucket)
	    throws IOException {
	PrintWriter writer = null;
	BufferedReader reader = null;
	try {
	    reader = new BufferedReader(new StringReader(newContents));
	    writer = new PrintWriter(new OutputStreamWriter(
		    dataBucket.getOutputStream(), "UTF-8"));
	    String line = reader.readLine();
	    while (line != null) {
		writer.print(line + "\r\n");
		line = reader.readLine();
	    }
	} finally {
	    try {
		if (writer != null) {
		    writer.close();
		}
	    } finally {
		if (reader != null) {
		    reader.close();
		}
	    }
	}
    }

    private void writeHeaders(FileBucket metaBucket) throws IOException {
	PrintWriter writer = null;
	try {
	    writer = new PrintWriter(new OutputStreamWriter(
		    metaBucket.getOutputStream(), "UTF-8"));
	    writer.print("Content-Type=text/x-freewiki-node-contents\r\n");
	    writer.print("X-FreeWiki-Node-Date=" + new Date().getTime() + "\r\n");
	} finally {
	    if (writer != null) {
		writer.close();
	    }
	}
    }

    private void insertKey(String key, FileBucket metaBucket, FileBucket dataBucket,
	    FileBucket ctBucket)
	    throws IOException, CollisionException, WikiStoreException {
	System.out.println("INSERTING KEY: " + key);
	try {
	    InsertRequest request = new InsertRequest(htl, key, null, metaBucket,
						      dataBucket, ctBucket);
	    ClientEventListenerImpl l = new ClientEventListenerImpl();
	    request.addEventListener(l);
	    clientFactory.obtainClient(request).execute();
	    l.waitUntilFinished();
	    if (l.isCollision()) {
		throw new CollisionException();
	    } else if (l.isFailed()) {
		throw new WikiStoreException("Insert key " + key + " failed.");
	    }
	} catch (java.net.MalformedURLException ex) {
	    throw new WikiStoreException("Cannot insert key: " + ex.toString());
	} catch (InsertSizeException ex) {
	    throw new WikiStoreException("Cannot insert key: " + ex);
	}
    }

    public FreenetWikiManager(String nodeAddress, int htl) {
	try {
	    this.htl = htl;
	    ClientCore core = ClientUtil.getTCPCore(0);
	    clientFactory = new FNPClient(core, ClientUtil.getAddress(nodeAddress));
	} catch (java.net.BindException ex) {
	    ex.printStackTrace();
	    throw new RuntimeException(ex.toString());
	} catch (Freenet.BadAddressException ex) {
	    ex.printStackTrace();
	    throw new RuntimeException(ex.toString());
	}
    }

    public WikiNodeData getNode(String wikiName, String nodeName)
	    throws WikiStoreException {
	return getNodeRevision(wikiName, nodeName, -1);
    }

    public WikiNodeData editNode(String wikiName, String nodeName, String newContents,
	    int mostRecentRevision) throws CollisionException, WikiStoreException {
	FileBucket metaBucket = null;
	FileBucket dataBucket = null;
	FileBucket ctBucket = null;
	try {
	    metaBucket = new FileBucket();
	    dataBucket = new FileBucket();
	    ctBucket = new FileBucket();
	    writeContents(newContents, dataBucket);
	    writeHeaders(metaBucket);
	    insertKey(nodeKey(wikiName, nodeName, mostRecentRevision + 1), metaBucket,
		      dataBucket, ctBucket);
	    WikiNodeData newData = getNodeRevision(wikiName, nodeName,
						   mostRecentRevision + 1);
	    if (newData.getRevisionNumber() != mostRecentRevision + 1) {
		throw new WikiStoreException("Cannot get inserted revision.");
	    }
	    return newData;
	} catch (IOException ex) {
	    throw new WikiStoreException(ex.toString());
	} finally {
	    cleanupBuckets(dataBucket, metaBucket, ctBucket);
	}
    }
}
